use beteran_common_rust as bcr;
use beteran_protobuf_rust as bpr;
use prost::Message;

///
pub struct Service {
  connection_broker: nats::asynk::Connection,
  queue_broker: String,
}

impl std::fmt::Debug for Service {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("beteran-frontend-server-edge::member_bank_withdraw::service::Service")
      .finish()
  }
}

impl Service {
  ///
  pub fn new(connection_broker: nats::asynk::Connection, queue_broker: String) -> Service {
    Service {
      connection_broker,
      queue_broker,
    }
  }

  pub async fn subscribe(&self) -> std::result::Result<(), std::boxed::Box<dyn std::error::Error>> {
    futures::try_join!(
      self.create_member_bank_withdraw(),
      self.list_member_bank_withdraws(),
      self.get_member_bank_withdraw(),
      self.update_member_bank_withdraw(),
      self.delete_member_bank_withdraw(),
    )
    .map(|_| ())
  }

  fn get_client_in_header(
    &self,
    message: &nats::asynk::Message,
  ) -> Result<bpr::models::core::network::Client, bcr::error::rpc::Error> {
    match &message.headers {
      Some(headers) => {
        let client = match headers.get(bpr::c2se::core::network::HEADER_CLIENT) {
          Some(c) => {
            let msg = base64::decode(c).map_err(|e| {
              bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
                message: format!("invalid header: {}", e),
              })
            })?;
            bpr::models::core::network::Client::decode(msg.as_slice()).map_err(|e| {
              bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
                message: format!("invalid header: {}", e),
              })
            })?
          }
          None => {
            return Err(bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
              message: "invalid client information".to_string(),
            }));
          }
        };

        Ok(client)
      }
      None => Err(bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
        message: "invalid header".to_string(),
      })),
    }
  }

  async fn get_member_session_with_renewal(
    &self,
    client: &bpr::models::core::network::Client,
  ) -> Result<Option<bpr::models::member_session::MemberSession>, bcr::error::rpc::Error> {
    let ss_get_member_session_req = bpr::ss::member_session::GetMemberSessionWithRenewalRequest {
      client: Some(client.clone()),
      request: Some(bpr::ss::member_session::get_member_session_with_renewal_request::Request {}),
    };

    let ss_get_member_session_res_msg = self
      .connection_broker
      .request(
        bpr::ss::member_session::SUBJECT_GET_MEMBER_SESSION_WITH_RENEWAL,
        ss_get_member_session_req.encode_to_vec(),
      )
      .await
      .map_err(|e| {
        bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
          code: bpr::protobuf::rpc::Error::SERVER_00,
          message: format!("server {}", e),
          data: None,
        })
      })?;

    let ss_get_member_session_res =
      bpr::ss::member_session::GetMemberSessionWithRenewalResponse::decode(
        ss_get_member_session_res_msg.data.as_slice(),
      )
      .map_err(|e| {
        bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
          code: bpr::protobuf::rpc::Error::SERVER_00,
          message: format!("server {}", e),
          data: None,
        })
      })?;

    if let Some(e) = ss_get_member_session_res.error {
      return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
        code: bpr::protobuf::rpc::Error::SERVER_00,
        message: format!("server {}", e),
        data: None,
      }));
    }

    if let Some(r) = ss_get_member_session_res.result {
      return Ok(r.member_session);
    }

    Ok(None)
  }

  async fn create_member_bank_withdraw(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::member_bank_withdraw::SUBJECT_CREATE_MEMBER_BANK_WITHDRAW,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;
        let member_session = match self.get_member_session_with_renewal(&client).await? {
          Some(ms) => ms,
          None => {
            return Err(bcr::error::rpc::Error::InvalidRequest(
              bcr::error::rpc::InvalidRequest {
                message: "Unauthorized".to_string(),
              },
            ));
          }
        };

        let req = bpr::c2se::member_bank_withdraw::CreateMemberBankWithdrawRequest::decode(
          message.data.as_slice(),
        )
        .map_err(|e| {
          bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
            message: format!("invalid request: {}", e),
          })
        })?;

        let ss_create_member_bank_withdraw_req =
          bpr::ss::member_bank_withdraw::CreateMemberBankWithdrawRequest {
            client: Some(client),
            request: Some(
              bpr::ss::member_bank_withdraw::create_member_bank_withdraw_request::Request {
                member_id: member_session.member_id.clone(),
                bank_name: req.bank_name,
                account_number: req.account_number,
                amount: req.amount,
                password: req.password,
                name: req.name,
                memo: req.memo,
              },
            ),
          };

        let ss_create_member_bank_withdraw_res_msg = self
          .connection_broker
          .request(
            bpr::ss::member_bank_withdraw::SUBJECT_CREATE_MEMBER_BANK_WITHDRAW,
            ss_create_member_bank_withdraw_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        let ss_create_member_bank_withdraw_res =
          bpr::ss::member_bank_withdraw::CreateMemberBankWithdrawResponse::decode(
            ss_create_member_bank_withdraw_res_msg.data.as_slice(),
          )
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        if let Some(e) = ss_create_member_bank_withdraw_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          }));
        }

        if let Some(r) = ss_create_member_bank_withdraw_res.result {
          message
            .respond(
              bpr::c2se::member_bank_withdraw::CreateMemberBankWithdrawResponse {
                error: None,
                result: Some(
                  bpr::c2se::member_bank_withdraw::create_member_bank_withdraw_response::Result {
                    member_bank_withdraw: r.member_bank_withdraw,
                  },
                ),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                code: bpr::protobuf::rpc::Error::SERVER_00,
                message: format!("server {}", e),
                data: None,
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::member_bank_withdraw::CreateMemberBankWithdrawResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn list_member_bank_withdraws(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::member_bank_withdraw::SUBJECT_LIST_MEMBER_BANK_WITHDRAWS,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;
        let member_session = match self.get_member_session_with_renewal(&client).await? {
          Some(ms) => ms,
          None => {
            return Err(bcr::error::rpc::Error::InvalidRequest(
              bcr::error::rpc::InvalidRequest {
                message: "Unauthorized".to_string(),
              },
            ));
          }
        };

        let req = bpr::c2se::member_bank_withdraw::ListMemberBankWithdrawsRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let ss_list_member_bank_withdraws_req = bpr::ss::member_bank_withdraw::ListMemberBankWithdrawsRequest {
          client: Some(client),
          request: Some(bpr::ss::member_bank_withdraw::list_member_bank_withdraws_request::Request{
            pagination: req.pagination,
            sorts: req.sorts,
            search: match req.search {
              Some(s) => Some(bpr::ss::member_bank_withdraw::list_member_bank_withdraws_request::request::Search {
                name_like: s.name_like,
                member_id: Some(member_session.member_id.clone()),
                bank_name_like: s.bank_name_like,
                account_number_like: s.account_number_like,
                amount: s.amount,
                state: s.state,
                memo_like: s.memo_like,
              }),
              None => None,
              }
          })
          
        };

        let ss_list_member_bank_withdraws_res_msg = self
          .connection_broker
          .request(
            bpr::ss::member_bank_withdraw::SUBJECT_LIST_MEMBER_BANK_WITHDRAWS,
            ss_list_member_bank_withdraws_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        let ss_list_member_bank_withdraws_res =
          bpr::ss::member_bank_withdraw::ListMemberBankWithdrawsResponse::decode(ss_list_member_bank_withdraws_res_msg.data.as_slice())
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                code: bpr::protobuf::rpc::Error::SERVER_00,
                message: format!("server {}", e),
                data: None,
              })
            })?;

        if let Some(e) = ss_list_member_bank_withdraws_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          }));
        }

        if let Some(r) = ss_list_member_bank_withdraws_res.result {
          message
            .respond(
              bpr::c2se::member_bank_withdraw::ListMemberBankWithdrawsResponse {
                error: None,
                result: Some(bpr::c2se::member_bank_withdraw::list_member_bank_withdraws_response::Result {
                  member_bank_withdraws: r.member_bank_withdraws,
                }),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                code: bpr::protobuf::rpc::Error::SERVER_00,
                message: format!("server {}", e),
                data: None,
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::member_bank_withdraw::ListMemberBankWithdrawsResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn get_member_bank_withdraw(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::member_bank_withdraw::SUBJECT_GET_MEMBER_BANK_WITHDRAW,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;
        let member_session = match self.get_member_session_with_renewal(&client).await? {
          Some(ms) => ms,
          None => {
            return Err(bcr::error::rpc::Error::InvalidRequest(
              bcr::error::rpc::InvalidRequest {
                message: "Unauthorized".to_string(),
              },
            ));
          }
        };

        let req = bpr::c2se::member_bank_withdraw::GetMemberBankWithdrawRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let ss_get_member_bank_withdraw_req = bpr::ss::member_bank_withdraw::GetMemberBankWithdrawRequest {
          client: Some(client),
          request: Some(bpr::ss::member_bank_withdraw::get_member_bank_withdraw_request::Request{
            id: req.id,  
          })
        };

        let ss_get_member_bank_withdraw_res_msg = self
          .connection_broker
          .request(
            bpr::ss::member_bank_withdraw::SUBJECT_GET_MEMBER_BANK_WITHDRAW,
            ss_get_member_bank_withdraw_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        let ss_get_member_bank_withdraw_res =
          bpr::ss::member_bank_withdraw::GetMemberBankWithdrawResponse::decode(ss_get_member_bank_withdraw_res_msg.data.as_slice())
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                code: bpr::protobuf::rpc::Error::SERVER_00,
                message: format!("server {}", e),
                data: None,
              })
            })?;

        if let Some(e) = ss_get_member_bank_withdraw_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          }));
        }

        if let Some(r) = ss_get_member_bank_withdraw_res.result {
          message
            .respond(
              bpr::c2se::member_bank_withdraw::GetMemberBankWithdrawResponse {
                error: None,
                result: Some(bpr::c2se::member_bank_withdraw::get_member_bank_withdraw_response::Result { member_bank_withdraw: r.member_bank_withdraw }),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                code: bpr::protobuf::rpc::Error::SERVER_00,
                message: format!("server {}", e),
                data: None,
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::member_bank_withdraw::GetMemberBankWithdrawResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn update_member_bank_withdraw(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::member_bank_withdraw::SUBJECT_UPDATE_MEMBER_BANK_WITHDRAW,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;
        let member_session = match self.get_member_session_with_renewal(&client).await? {
          Some(ms) => ms,
          None => {
            return Err(bcr::error::rpc::Error::InvalidRequest(
              bcr::error::rpc::InvalidRequest {
                message: "Unauthorized".to_string(),
              },
            ));
          }
        };

        let req = bpr::c2se::member_bank_withdraw::UpdateMemberBankWithdrawRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let ss_update_member_bank_withdraw_req = bpr::ss::member_bank_withdraw::UpdateMemberBankWithdrawRequest {
          client: Some(client),
          request: Some(bpr::ss::member_bank_withdraw::update_member_bank_withdraw_request::Request{
            id: req.id,
            bank_name: req.bank_name,
 account_number: req.account_number,
amount: req.amount,
password: req.password,
            name: req.name,
            memo: req.memo,
          })
        };

        let ss_update_member_bank_withdraw_res_msg = self
          .connection_broker
          .request(
            bpr::ss::member_bank_withdraw::SUBJECT_UPDATE_MEMBER_BANK_WITHDRAW,
            ss_update_member_bank_withdraw_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        let ss_update_member_bank_withdraw_res = bpr::ss::member_bank_withdraw::UpdateMemberBankWithdrawResponse::decode(
          ss_update_member_bank_withdraw_res_msg.data.as_slice(),
        )
        .map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          })
        })?;

        if let Some(e) = ss_update_member_bank_withdraw_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          }));
        }

        if let Some(r) = ss_update_member_bank_withdraw_res.result {
          message
            .respond(
              bpr::c2se::member_bank_withdraw::UpdateMemberBankWithdrawResponse {
                error: None,
                result: Some(bpr::c2se::member_bank_withdraw::update_member_bank_withdraw_response::Result {
                  member_bank_withdraw: r.member_bank_withdraw,
                }),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                code: bpr::protobuf::rpc::Error::SERVER_00,
                message: format!("server {}", e),
                data: None,
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::member_bank_withdraw::UpdateMemberBankWithdrawResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn delete_member_bank_withdraw(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::member_bank_withdraw::SUBJECT_DELETE_MEMBER_BANK_WITHDRAW,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;
        let member_session = match self.get_member_session_with_renewal(&client).await? {
          Some(ms) => ms,
          None => {
            return Err(bcr::error::rpc::Error::InvalidRequest(
              bcr::error::rpc::InvalidRequest {
                message: "Unauthorized".to_string(),
              },
            ));
          }
        };

        let req = bpr::c2se::member_bank_withdraw::DeleteMemberBankWithdrawRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let ss_delete_member_bank_withdraw_req = bpr::ss::member_bank_withdraw::DeleteMemberBankWithdrawRequest {
          client: Some(client),
          request: Some(bpr::ss::member_bank_withdraw::delete_member_bank_withdraw_request::Request{
            id: req.id,
          })
        };

        let ss_delete_member_bank_withdraw_res_msg = self
          .connection_broker
          .request(
            bpr::ss::member_bank_withdraw::SUBJECT_DELETE_MEMBER_BANK_WITHDRAW,
            ss_delete_member_bank_withdraw_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        let ss_delete_member_bank_withdraw_res = bpr::ss::member_bank_withdraw::DeleteMemberBankWithdrawResponse::decode(
          ss_delete_member_bank_withdraw_res_msg.data.as_slice(),
        )
        .map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          })
        })?;

        if let Some(e) = ss_delete_member_bank_withdraw_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          }));
        }

        if let Some(r) = ss_delete_member_bank_withdraw_res.result {
          message
            .respond(
              bpr::c2se::member_bank_withdraw::DeleteMemberBankWithdrawResponse {
                error: None,
                result: Some(bpr::c2se::member_bank_withdraw::delete_member_bank_withdraw_response::Result {}),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                code: bpr::protobuf::rpc::Error::SERVER_00,
                message: format!("server {}", e),
                data: None,
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::member_bank_withdraw::DeleteMemberBankWithdrawResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }
}
